package jsetl.test;

import jsetl.*;
import jsetl.exception.Failure;
import jsetl.test.utilities.comparing.ConstraintStringComparator;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.jupiter.api.Assertions.*;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Scanner;

public class SpeTests {
    // STATIC MEMBERS
    static private ByteArrayOutputStream byteArrayOutputStream;
    static private PrintStream oldPrintStream;
    static private Solver solver;
    
    // INITIALIZATION AND TERMINATION BEFORE AND AFTER EVERY TEST
    @Before
    public void initialize() {
        byteArrayOutputStream = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(byteArrayOutputStream);
        oldPrintStream = System.out;
        System.setOut(ps);
        solver = new Solver();
    }

    @After
    public void terminate() throws IOException {
        byteArrayOutputStream.close();
        System.out.close();
        System.setOut(oldPrintStream);
    }
    
    
    // TEST EXAMPLE 06 CONSTRAINT SOLVING
    @Test
    public void testEs06ConstraintSolving()throws Failure {
        LSet s1 = LSet.empty().ins(new LVar("z")).ins(1);  // the set \{1,z\}
        LVar x = new LVar("x"), y = new LVar("y");
        LSet s2 = LSet.empty().ins(y).ins(x);  // the set \{x,y\}
        Solver solver = new Solver();
        solver.add(s1.eq(s2).and(x.neq(1)));   // the constraint s1=s2 \(\land\) x\(\neq\)1
        solver.solve();
        y.output();
        solver.showStore();

        assertEquals("y = 1\r\n" +
                "Store: _x neq 1\r\n", byteArrayOutputStream.toString());
    }
    
    
    // TEST EXAMPLE 07 ALL PERMUTATIONS
    public static void allPermutations(Integer[] A) {
        LSet E = LSet.empty().insAll(A);
        LSet S = LSet.mkSet(A.length);
        //LSet S = LSet.mkSet(E.getSize()); //This way it ignores repetitions in A
        solver.add(E.eq(S));
        solver.forEachSolution(i -> {System.out.print(i+")  ");
            S.printElems(' ');});
    }

    @Test
    public void testEs07AllPermutations() {
        Integer[] elems = {1,2,3};
        allPermutations(elems);

        assertEquals("1)  1 2 3\r\n" +
                "2)  1 3 2\r\n" +
                "3)  2 1 3\r\n" +
                "4)  2 3 1\r\n" +
                "5)  3 1 2\r\n" +
                "6)  3 2 1\r\n", byteArrayOutputStream.toString());
    }
    
    
    // TEST EXAMPLE 08 RIS OBJECT CREATION
    @Test
    public void testEs08RisObjectCreation(){
        IntLVar x = new IntLVar();
        IntLSet d = new IntLSet(-2,2);
        Constraint f = x.mod(2).eq(0);
        Ris ris = new Ris(x,d,f);   // \{x : [-2,2] \(|\) x mod 2 = 0 \(@\) x\}
        ris.output();

        ConstraintStringComparator constraintStringComparator =
                new ConstraintStringComparator("_N2 = { __N1 : {-2..2} | __I0 = 0 AND __I0 = __N1 mod 2 @ __N1 }\r\n",
                        byteArrayOutputStream.toString());
        assertTrue(constraintStringComparator.areEqualInSameOrder());
    }
    
    
    // TEST EXAMPLE 09 RIS EXPAND
    @Test
    public void testEs09RisExpand(){
        IntLVar x = new IntLVar();
        IntLSet d = new IntLSet(-2,2);
        Constraint f = x.mod(2).eq(0);
        Ris ris = new Ris(x,d,f);   // \{x : [-2,2] \(|\) x mod 2 = 0 \(@\) x\}

        LSet S = ris.expand().setName("S");
        S.output();

        assertEquals("S = {0,-2,2}\r\n", byteArrayOutputStream.toString());
    }
    
    
    // TEST EXAMPLE 10 RIS OBJECT CREATION
    @Test
    public void testEs10aRisObjectCreation(){
        LSet S = new LSet();
        LSet s1 = LSet.empty().ins(2).ins(0);
        LSet s2 = LSet.empty().ins(2);
        LSet s3 = LSet.empty().ins(2).ins(0).ins(1);
        LSet D = LSet.empty().ins(s1).ins(s2).ins(s3);
        LSet A = LSet.empty().ins(0);

        Ris ris = new Ris(S,D,A.subset(S));   // {S : D | subset(A,S) @ S}

        LSet R = ris.expand().setName("R");
        R.output();

        assertEquals("R = {{1,0,2},{0,2}}\r\n", byteArrayOutputStream.toString());
    }

    @Test
    public void testEs10bRisObjectCreation(){
        LSet S = new LSet();
        IntLVar c = new IntLVar();
        LPair p1 = new LPair(LSet.empty().ins(2).ins(0),2);
        LPair p2 = new LPair(LSet.empty(),0);
        LPair p3 = new LPair(LSet.empty().ins(0).ins(0),2);
        LPair p4 = new LPair(LSet.empty().ins(2),1);
        LPair p5 = new LPair(LSet.empty().ins(1,2,3,3,2,1),3);
        LPair p6 = new LPair(LSet.empty().ins(1,2,3,3,2,1),4);
        LSet D = LSet.empty().ins(p1).ins(p2).ins(p3).ins(p4).ins(p5).ins(p6);
        
        Ris ris = new Ris(new LPair(S,c),D,S.size(c).and(c.gt(1)));   // {P:D|size(S,c) & c > 1}

        LSet R = ris.expand().setName("R");
        R.output();

        assertEquals("R = {[{1,2,3},3],[{0,2},2]}\r\n", byteArrayOutputStream.toString());
    }
    
    
    // TEST EXAMPLE 11 RIS CONSTRAINTS
    @Test
    public void testEs11RisConstraintsA (){
        IntLVar x = new IntLVar("x");
        IntLSet d = new IntLSet(-2,2);
        LSet s = LSet.empty().ins(2).ins(0).ins(-2);
        Constraint f = x.mod(2).eq(0);
        Ris ris = new Ris(x,d,f);
        solver.add(ris.eq(s));
        LVar y = new LVar(1);
        solver.add(y.nin(ris));

        assertTrue(solver.check());

    }

    @Test
    public void testEs11RIsConstraintsB (){
        IntLVar x = new IntLVar("x");
        IntLSet d = new IntLSet(-2,2);
        LSet s = LSet.empty().ins(2).ins(0).ins(-2);
        Constraint f = x.mod(2).eq(0);
        Ris ris = new Ris(x,d,f);
        LVar y = new LVar(1);
        solver.add(ris.eq(s).and(y.nin(ris)));

        assertTrue(solver.check());

    }
    
    
    // TEST EXAMPLE 12 RIS CONSTRAINTS
    @Test
    public void testEs12ARisConstraints () throws Failure {
        IntLVar x = new IntLVar("x");
        LSet d = new LSet("d");
        Ris ris = new Ris(x,d,x.mod(2).eq(0));
        LSet s = LSet.empty().ins(6).ins(4).ins(2); // s = {2,4,6}
        solver.solve(s.eq(ris));
        d.output();
        solver.showStore();

        assertEquals("d = {2,4,6|_N1}\r\n" +
                "Store: { _x : _N1 | __I0 = 0 AND __I0 = _x mod 2 @ _x } = {}\r\n", byteArrayOutputStream.toString());

    }
    
    @Test
    public void testEs12BRisConstraints () {
        LSet a = new LSet(), b = new LSet(), c = new LSet();
        IntLVar x = new IntLVar();
        LSet d = new LSet();
        solver.add(a.eq(new Ris(x,d,x.neq(0))));
        solver.add(a.union(b,c).and(a.disj(c)));
        solver.add(a.neq(LSet.empty()));
        assertThrows(Failure.class, ()->{solver.solve();});
    }
    
    
    // TEST EXAMPLE 13 MIN
    public static LVar minValue(LSet S) throws Failure {
        IntLVar x = new IntLVar(), m = new IntLVar();
        Ris ris = new Ris(x,S,m.le(x));
        solver.add(m.in(S).and(S.subset(ris)));
        solver.solve();
        return m;
    }

    @Test
    public void testEs13MinA() throws Failure{
        Integer[] sampleSetElems = {8,4,6,2,10,5};
        LSet A = LSet.empty().insAll(sampleSetElems);
        LVar min = minValue(A).setName("min");
        min.output();

        assertEquals("min = 2\r\n", byteArrayOutputStream.toString());
    }

    @Test
    public void testEs13MinB() throws Failure{
        IntLVar z = new IntLVar("z");
        LSet A = LSet.empty().ins(8,z,4,6);
        LVar min = minValue(A).setName("min");
        do{
            min.output();
            z.output();
            System.out.println("");
        }while(solver.nextSolution());
        assertEquals("min = _z\r\n" +
                "z = unknown, z =< 4\r\n" +
                "\r\n" +
                "min = 4\r\n" +
                "z = unknown, z >= 4\r\n\r\n", byteArrayOutputStream.toString());
    }


    // TEST EXAMPLE 14 MAP COLORING
    public static void coloring(LSet regions, LSet map, LSet colors) throws Failure {
        solver.add(regions.subset(colors));
        LSet p = new LSet();
        Ris ris = new Ris(p, map, p.size(2));
        solver.add(map.subset(ris));
        solver.solve();
    }
    @Test
    public void testEs14MapColoring() throws Failure{
        LVar r1 = new LVar("r1"), r2 = new LVar("r2"), r3 = new LVar("r3");
        LSet regions = LSet.empty().ins(r3).ins(r2).ins(r1);
        LSet map = LSet.empty().ins(LSet.empty().ins(r2).ins(r3)).ins(LSet.empty().ins(r1).ins(r2));
        LSet colors = LSet.empty().ins("blue").ins("red");

        coloring(regions,map,colors);

        r1.output();
        r2.output();
        r3.output();

        assertTrue(solver.nextSolution());
        r1.output();
        r2.output();
        r3.output();

        assertFalse(solver.nextSolution());
        assertEquals("r1 = red\r\n" +
                "r2 = blue\r\n" +
                "r3 = red\r\n" +
                "r1 = blue\r\n" +
                "r2 = red\r\n" +
                "r3 = blue\r\n", byteArrayOutputStream.toString());
    }

    // TEST EXAMPLE 15 IS PRIME
    public static Boolean isPrime(int n) {
        if (n <= 0) return false;
        IntLVar N = new IntLVar(n);
        IntLVar x = new IntLVar();
        IntLSet D = new IntLSet(2,n/2);
        //IntLSet D = new IntLSet(2,(int)Math.sqrt(n));
        Ris ris = new Ris(x,D,N.mod(x).eq(0));
        return solver.check(ris.eq(LSet.empty()));
    }

    public static boolean isPrimeJava(int n){
        if(n <= 1) return false;
        if(n == 2) return true;
        if(n % 2 == 0)
            return false;
        for(int i = 3; i < n/2; i = i + 2)
            if(n % i == 0)
                return false;
        return true;
    }

    @Test
    public void testEs15IsPrime () {
        for(int i = 2; i < 200; ++i) // with i in [2,200) works
            assertEquals(isPrimeJava(i), isPrime(i));
        fail("for 1 it doesn't work currently");
    }


    // TEST EXAMPLE 16 SQUARE FUNCTION
    @Test
    public void testEs16SquareFunctionA () throws Failure{
        int n = 5;

        IntLVar x = new IntLVar();
        LSet D = new LSet();
        Ris sqr = new Ris(x,D,Constraint.truec(),new LPair(x,x.mul(x)));

        IntLVar y = new IntLVar("y");
        solver.solve(new LPair(n,y).in(sqr));
        y.output();

        assertFalse(solver.nextSolution());
        assertEquals("y = 25\r\n", byteArrayOutputStream.toString());
    }

    @Test
    public void testEs16SquareFunctionB () throws Failure{
        int n = 25;

        IntLVar x = new IntLVar();
        LSet D = new LSet();
        Ris sqr = new Ris(x,D,Constraint.truec(),new LPair(x,x.mul(x)));

        IntLVar y = new IntLVar("y");
        solver.solve(new LPair(y,n).in(sqr));
        y.output();

        assertFalse(solver.nextSolution());
        assertEquals("y = unknown -- Domain: {-5, 5}\r\n", byteArrayOutputStream.toString());
    }


    // TEST EXAMPLE 17 MAP LIST
    public static void mapList(Integer[] A,LSet f) throws Failure {
        for(int i=0; i<A.length; i++) {
            IntLVar y = new IntLVar();
            solver.solve(new LPair(A[i],y).in(f));
            A[i] = y.getValue();
        }
    }
    @Test
    public void testEs17MapList () throws Failure{
        IntLVar x = new IntLVar();
        LSet D = new LSet();
        Ris sqr = new Ris(x,D,Constraint.truec(),new LPair(x,x.mul(x)));
        Integer[] elems = {3,5,7};
        mapList(elems,sqr);
        System.out.println(Arrays.asList(elems));

        assertEquals("[9, 25, 49]\r\n", byteArrayOutputStream.toString());
    }

    public static void mapListVariant(Integer[] A,LSet f) throws Failure {
        for(int i=0; i<A.length; i++) {
            IntLVar y = new IntLVar();
            if(solver.check(new LPair(y,A[i]).in(f)))
                System.out.println(A[i]);
        }
    }

    @Test
    public void testEs17MapListVariant () throws Failure{
        IntLVar x = new IntLVar();
        LSet D = new LSet();
        Ris sqr = new Ris(x,D,Constraint.truec(),new LPair(x,x.mul(x)));
        Integer[] elems = {3,100,5,81,7,36};
        mapListVariant(elems,sqr);

        assertEquals("100\r\n81\r\n36\r\n", byteArrayOutputStream.toString());
    }


    // TEST EXAMPLE 18 THEOREM PROVER
    @Test
    public void testEs18TheoremProver(){
        LSet A = new LSet(), B = new LSet(), C = new LSet();
        solver.add(A.inters(B,C));      // the constraint C = A cap B
        LVar x = new LVar();
        Ris D = new Ris(x,A,x.in(B)).setName("D");   // D = {X:A | X in B}
        solver.add(D.neq(C));           // the constraint C neq D

        assertThrows(Failure.class, solver::solve);
    }


    // TEST EXAMPLE 19 THEOREM PROVER 2
    @Test
    public void testEs19TheoremProver(){
        LSet S = new LSet();
        IntLVar x = new IntLVar();
        Ris ris = new Ris(x,S,x.gt(0));
        Constraint c = S.subset(ris).and(S.contains(-1));


        assertThrows(Failure.class, ()->{solver.solve(c);});
    }


    // TEST EXAMPLE 20 SQUARES
    @Test
    public void testEs20Squares(){
        IntLVar x = new IntLVar();     //ris = {x:[1,10] | x mod 2=0 @ x*x}
        IntLSet D = new IntLSet(1,10);
        Ris ris = new Ris(x,D,x.mod(2).eq(0),x.mul(x));
        LSet Sqrs = ris.expand();

        assertEquals("{4,16,36,64,100}", Sqrs.toString());
    }


    // TEST EXAMPLE 21 DUMMY
    @Test
    public void testEs21Dummy() throws Failure{
        LSet R = new LSet(), D = new LSet();
        LVar x = new LVar(), y = new LVar();
        Ris ris = new Ris(x,D,new LPair(x,y).in(R),new LPair(x,y),y);

        solver.solve(new LPair(1,2).in(ris).and(new LPair(3,4).in(ris)));
        D.output(); R.output();

        assertEquals("", byteArrayOutputStream.toString());
    }


    // TEST EXAMPLE 22 FACTORIAL
    @Test
    public void testEs22FactorialOLD() throws Failure{
        LSet fact=new LSet();
        IntLVar x=new IntLVar();
        IntLVar pred_x=new IntLVar(), z=new IntLVar(), fact_x=new IntLVar();
        Constraint c=x.gt(0).and(pred_x.eq(x.sub(1)))
                .and(new LPair(pred_x,z).in(fact))
                .and(fact_x.eq(z.mul(x)));
        Ris fact1=new Ris(x,new LSet(),c,new LPair(x,fact_x),pred_x,z,fact_x);
        solver.add(fact.eq(fact1.ins(new LPair(0,1))));

        IntLVar ff = new IntLVar("ff");
        solver.add(new LPair(5,ff).in(fact));
        solver.solve();
        assertEquals((Integer) 120, ff.getValue());
        ff.output();

        solver.clearStore();
        IntLVar n = new IntLVar("n");
        solver.add(new LPair(n,120).in(fact));
        solver.solve();
        assertEquals((Integer) 5, n.getValue());
        n.output();

        assertEquals("ff = 120\r\n" +
                "n = 5\r\n", byteArrayOutputStream.toString());
    }
    @Test
    public void testEs22FactorialNEW() throws Failure{
        LSet fact=new LSet();
        IntLVar x = new IntLVar();
        IntLVar z = new IntLVar();
        Constraint c = x.gt(0).and(new LPair(x.sub(1),z).in(fact));
        Ris fact1=new Ris(x,new LSet(),c,new LPair(x,z.mul(x)),z);
        solver.add(fact.eq(fact1.ins(new LPair(0,1))));

        IntLVar ff = new IntLVar("ff");
        solver.add(new LPair(5,ff).in(fact));
        solver.solve();
        assertEquals((Integer) 120, ff.getValue());
        ff.output();

        solver.clearStore();
        IntLVar n = new IntLVar("n");
        solver.add(new LPair(n,120).in(fact));
        solver.solve();
        assertEquals((Integer) 5, n.getValue());
        n.output();

        assertEquals("ff = 120\r\n" +
                "n = 5\r\n", byteArrayOutputStream.toString());
    }


    // TEST EXAMPLE 22 REACHABLE NODES
    public static LSet reachableNodesOLD(LSet nodes, LRel edges, LVar start){
        LVar x = new LVar("x");
        LVar y = new LVar("y");
        LSet reachable = new LSet();
        Ris ris = new Ris(x, nodes, x.eq(start)
                .or(new LPair(start,x).in(edges))
                .or(y.in(reachable).and(new LPair(y, x).in(edges))),x,
                y);
        reachable.eq(ris).check();
        return ris.expand();
    }
    @Test
    public void testEs23ReachableNodesOLD(){
        Integer[] nodesArray = {1,2,3,4,5,6};
        LPair[] edgesArray = {new LPair(1,2), new LPair(1,3), new LPair(2,5), new LPair(4,6)};
        LSet nodes = LSet.empty().insAll(nodesArray);
        LRel edges = LRel.empty().insAll(edgesArray);

        LVar start = new LVar(1);

        reachableNodesOLD(nodes, edges, start).setName("reachable").output();

        assertEquals("reachable = {1,2,3,5}\r\n", byteArrayOutputStream.toString());
    }

    public static LSet reachableNodes(LSet nodes, LRel edges, LVar start){
        LVar x = new LVar("x");
        LVar y = new LVar("y");
        LSet reachable = new LSet();
        Ris ris = new Ris(x, nodes, x.eq(start)
                .or(y.in(reachable).and(new LPair(y, x).in(edges))),x,
                y);
        reachable.eq(ris).check();
        return ris.expand();
    }
    @Test
    public void testEs23ReachableNodesNew(){
        LSet nodes = LSet.empty().ins(1,2,3,4,5,6);
        LRel edges = LRel.empty().ins(new LPair(1,2), new LPair(1,3), new LPair(2,5), new LPair(4,6));
        LVar start = new LVar(1);

        reachableNodes(nodes, edges, start).setName("reachable").output();

        assertEquals("reachable = {1,2,3,5}\r\n", byteArrayOutputStream.toString());
    }








}
