package jsetl.test.utilities.comparing;

import jsetl.Constraint;

import java.util.HashMap;
import java.util.Map;

/**
 * Class used to normalize strings representing constraint conjunctions.
 * Normalization consists on replacing each (automatically generated) number in variable names with
 * new numbers (starting from 1) in a consistent way (i.e. each occurrence of the same number is replaced with the
 * same new number).
 * The application of this algorithm ensures that two strings will be mapped to the same normalized string
 * if and only if the two strings differ only in variable name numbers (and in a consistent way).
 * This class is used inside <code>ConstraintStringComparator</code> before the string comparison occurs.
 * @see ConstraintStringComparator
 * @author Andrea Fois
 */
public class ConstraintStringNormalizer{
    /**
     * Character used to recognized variable name prefixes in constraint string representations.
     */
    private char variableNamePrefix;

    /**
     * Data structure used to store generated associations between numbers appearing in the given string
     * and the new generated numbers
     */
    private Map<Integer, Integer> translationMap;

    /**
     * Reference to the string that should be normalized
     */
    private String constraintString;

    /**
     * This integer is used and incremented to generate new values for the numbers in variable names.
     */
    private int _ID;

    /**
     * Constructs a new normalizer for the given string using the given prefix.
     * @param variableNamePrefix prefix used in <code>constraintString</code> before variable names
     * @param constraintString string to normalize
     */
    public ConstraintStringNormalizer(char variableNamePrefix, String constraintString){
        this.variableNamePrefix = variableNamePrefix;
        this.constraintString = constraintString;
        translationMap = new HashMap<>();
    }

    /**
     * Constructs a new normalizer for the given string using the default prefix.
     * @param constraintString string to normalize
     * @see ConstraintStringComparator#DEFAULT_VARIABLE_NAME_PREFIX
     */
    public ConstraintStringNormalizer(String constraintString){
        this(ConstraintStringComparator.DEFAULT_VARIABLE_NAME_PREFIX, constraintString);
    }

    /**
     * Constructs a new normalizer for the given constraint using the given prefix.
     * Automatically gets the string corresponding to the given constraint by using its <code>toString()</code> method.
     * @param variableNamePrefix prefix used in <code>constraintString</code> before variable names
     * @param constraint constraint whose string representation is to normalize
     */
    public ConstraintStringNormalizer(char variableNamePrefix, Constraint constraint){
        this(variableNamePrefix, constraint.toString());
    }

    /**
     * Constructs a new normalizer for the given constraint using the default prefix.
     * Automatically gets the string corresponding to the given constraint by using its <code>toString()</code> method.
     * @param constraint constraint whose string representation is to normalize
     * @see ConstraintStringComparator#DEFAULT_VARIABLE_NAME_PREFIX
     */
    public ConstraintStringNormalizer(Constraint constraint){
        this(constraint.toString());
    }

    /**
     * Returns the number present in the <code>constraintString</code> starting from <code>start</code>.
     * NOTE that this method assumes that the character at index <code>start</code> is a digit and so does not check if it actually is.
     * @param start index of the first digit in the <code>constraintString</code>
     * @return the number found, 0 if no number was found.
     */
    private int getNumber(int start){
        int i = start;
        int number = 0;
        while(i < constraintString.length()){
            char character = constraintString.charAt(i);
            if(character < '0' || character > '9')
                break;

            int digit = Character.getNumericValue(character);
            number = 10 * number + digit;
            ++i;
        }

        return number;
    }

    /**
     * Resets the id generator.
     */
    private void resetID(){
        this._ID = 0;
    }

    /**
     * Returns a new numerical id, different from those generated before.
     * @return a numerical id
     */
    private int getID(){
        return ++this._ID;
    }

    /**
     * Constructs and returns the normalized version of <code>constraintString</code>.
     * Normalization leaves the string as it is except for the numbers in variable names, which are replaced
     * consistently with ids generated starting from zero.
     * The application of this algorithm ensures that two strings will be mapped to the same normalized string
     * if and only if the two strings differ only in variable name numbers (and in a consistent way).
     * @return <code>constraintString</code> normalized
     */
    public String normalize(){
        resetID();
        String newString = "";
        boolean variableNameFlag = false;
        for(int i = 0; i < constraintString.length(); ++i){
            char c = constraintString.charAt(i);
            if(variableNameFlag && Character.isDigit(c)){
                int number = getNumber(i);
                i += ("" + number).length() - 1;
                newString += getNumberTranslated(number);
                variableNameFlag = false;
            }
            else {
                newString += c;
                if (c == variableNamePrefix)
                    variableNameFlag = true;
                else if(!Character.isLetter(c))
                    variableNameFlag = false;
            }
        }

        return newString;
    }

    /**
     * Retrieves and returns the translated number, if already translated, otherwise
     * generates a new translation for it (using the id generator) and stores it for future
     * usages (other occcurrences of the same number in <code>constraintString</code>).
     * @param number the number to translate
     * @return the translation (aka new value for) <code>number</code>
     */
    private int getNumberTranslated(int number){
        if(translationMap.containsKey(number))
            return translationMap.get(number);

        int newNumber = getID();
        translationMap.put(number, newNumber);
        return newNumber;
    }
}
